//-----------------------------------------------------------------------------
// File: snfitter.C
// Description: Fit 3-parameter models to the latest compilation of Type Ia 
//              supernova data.
//
//              a    - universal scale factor
//
//              OM   - Omega_M
//              OL   - Omega_L
//              H    - related to Hubble constant in some complicated
//                     calibration-dependent way. This parameter is best
//                     regarded as a nuisance parameter
//
//              For some models OM and OL may have different meanings.
//
// Created: June 2008 HBP
// Updated for ESHEP 2012, La Pommeraye, Anjou, France
//-----------------------------------------------------------------------------
#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <cmath>

#include "TROOT.h"
#include "TMath.h"
#include "TStyle.h"
#include "TH1F.h"
#include "TGraphErrors.h"
#include "TCanvas.h"
#include "TMinuit.h"
#include "util.h"
using namespace std;
//-----------------------------------------------------------------------------
int NSN=0;              // Number of Type Ia supernovae in data-set
double Z[2000];         // redshift
double MU[2000];        // distance modulus
double MU_ERR[2000];    // distance modulus uncertainty
double MU_FIT[2000];    // fitted distance moduli
double MU_RES[2000];    // fit residuals (mu - mu-best-fit)
double OFFSET=25+5*TMath::Log10(2.99*pow(10.0, 5.0));
//-----------------------------------------------------------------------------

// Model with cosmological constant
double lambdaModel(double a, double OM, double OL)
{
  return OM + (1 - OM - OL)*a + OL*a*a*a;
}

// Dungan-Prosper Varying-G or phantom energy model
double phantomModel(double a, double OM, double OL)
{
  OL = OL; // some stupidity to avoid compiler warnings
  return exp(OM*(a-1));
}

// pointer to model
double (*model)(double, double, double);

double distanceModulus(double z, double OM, double OL, double H)
{
  int N=500;
  double F = 0;
  double a = 1.0/(1+z);
  double h = (1-a) / N;
  
  for(int i=0; i < N; i++)
    {
      double x = a + (i+0.5)*h;
      F = F + 1.0/sqrt(x * model(x, OM, OL));
    }

  F = F*h;
  double OK = 1 - OM - OL;
  double rootOK = sqrt(fabs(OK));
  if ( OK > 0 )
    F = sinh(rootOK * F)/ rootOK;
  else if ( OK < 0 )
    F = sin(rootOK * F)/ rootOK;

  return 5 * TMath::Log10((1+z)*F/H) + OFFSET;
}

double lifetime(double OM, double OL, double H)
{
  // Compute lifetime assuming a value of Hubble's constant

  int N=1000;
  double F = 0;
  double h = 1.0 / N;  
  for(int i=0; i < N; i++)
    {
      double x = (i+0.5)*h;
      F = F + sqrt(x / model(x, OM, OL));
    }
  return F/H;
}

// function to be minimized
static int COUNT = 0;
void chisquared(int&    npar,
		double* grad,
		double& fval,
		double* xval,
		int    iflag)
{
  npar = npar; // to avoid compiler warnings
  grad = grad;
  iflag= iflag;

  double OM=xval[0];
  double OL=xval[1];
  double H=xval[2];
  COUNT++;

  fval = 0;
  for(int i=0; i < NSN; i++)
    fval += pow((MU[i] - distanceModulus(Z[i], OM, OL, H))/MU_ERR[i], 2);
  if ( COUNT % 1 == 0 )
    {
      cout << COUNT << "\t" 
	   << fval << "\tOM = " << OM << "\tOL = " << OL << endl;
    }
}
//-----------------------------------------------------------------------------
void sfitter(int which=1)
{
  gROOT->ProcessLine(".L util.cc+");
  setStyle();

  string postfix("");
  // Choose model
  if ( which == 1)
    {
      model = lambdaModel;
      postfix = "_Lambda";
    }
  else
    {
      model = phantomModel;
      postfix = "_Phantom";
    }

  // Read in data into Z, MU and MU_ERR

  string namen = "SCPUnion2.1_mu_vs_z";
  string nom   = namen + postfix;

  ifstream input(string(namen+".txt").c_str());
  string line;

  // skip header
  for(int i=0; i < 5; i++)
    {
      getline(input, line);
      cout << line << endl;
    }

  // read data
  int ii = 0;
  double minz = 9999.0;
  double maxz =-1;
  double minmu=9999.0;
  double maxmu=-1;
  double tmp;
  while ( input >> line >> Z[ii] >> MU[ii] >> MU_ERR[ii] >> tmp )
    {
      if ( Z[ii] < minz ) minz = Z[ii];
      if ( Z[ii] > maxz ) maxz = Z[ii];
      if ( MU[ii] < minmu ) minmu = MU[ii];
      if ( MU[ii] > maxmu ) maxmu = MU[ii];
      cout << Z[ii] << "\t" << MU[ii] << "\t" << MU_ERR[ii] << endl;
      NSN++;
      ii++;
    }
  cout << "Number of Type Ia data points: " << NSN << endl;
  cout << "Z range:  " << minz << " ... " << maxz << endl;
  cout << "mu range: " << minmu << " ... " << maxmu << endl;

  //---------------------------------------------------------------------------
  // Plot data
  //---------------------------------------------------------------------------
  TCanvas* canvas = new TCanvas(string("fig_"+nom).c_str(), 
				"SN1a", 10, 10, 500, 500);

  TGraphErrors* g = mkgraphErrors(NSN, Z, MU, MU_ERR,
				  "redshift z", "distance modulus #mu", 
				  0, 1.5, 32.0, 48.0, kBlue+1);
  g->SetLineWidth(1);
  g->SetMarkerSize(0.2);

  canvas->cd();
  g->Draw("AP");

  Scribe* wrtext = new Scribe(0.25,0.45);

  wrtext->write("The Union2.1 Compilation");
  wrtext->write("The Supernova Cosmology Project");
  wrtext->write("http://supernova.lbl.gov/Union/figures/");
  wrtext->write("SCPUnion2.1_mu_vs_z.txt");
  canvas->SaveAs(".gif");
  canvas->SaveAs(".pdf");

  //---------------------------------------------------------------------------
  // Fit 
  //---------------------------------------------------------------------------
  int N=3;  // 3-parameter fit (one parameter is a nuisance parameter)

  vector<string> name(N);
  TMinuit minuit(N);
  minuit.SetFCN(chisquared);
  
  int flag;
  double OMguess = 0.4;
  double OMmax = 5.0;
  double OMmin = 0.0;
  double OMstep= (OMmax-OMmin)/200;
  name[0] = "OM";
  minuit.mnparm(0, name[0].c_str(), OMguess, OMstep, OMmin, OMmax, flag);

  double OLguess = 0.6;
  double OLmax = 5.0;
  double OLmin = 0.0;
  double OLstep= (OLmax-OLmin)/200;
  name[1] = "OL";
  minuit.mnparm(1, name[1].c_str(), OLguess, OLstep, OLmin, OLmax, flag);
  if ( which != 1) minuit.FixParameter(1);
	
  // In the context of this fit, H is best regarded as a nuisance parameter
  double Hguess = 70;
  double Hmax   = 500;
  double Hmin   =  20;
  double Hstep  = (Hmax-Hmin)/50;
  name[2] = "nuisance parameter";
  minuit.mnparm(2, name[2].c_str(), Hguess, Hstep, Hmin, Hmax, flag);

  // Now minimize chisquared

  cout << endl << "GO!" << endl << endl;
  minuit.SetErrorDef(0.5);
  minuit.Migrad();

  // Get parameters

  double fmin, fedm, errdef;
  int nvpar, nparx, status;
  minuit.mnstat(fmin, fedm, errdef, nvpar, nparx, status);
  double par[10], err[10];
  if ( status == 3 )
    {
      cout << endl << "minimization succeeded" << endl;
      for (int i=0; i < N; i++)
        {
          minuit.GetParameter(i, par[i], err[i]);
	  cout << i << "\t" << name[i] 
	       << "\t" << par[i] << " +/-" << err[i] << endl;
        }
    }
  else
    {
      cout << endl << "minimization FAILED" << endl;
      exit(0);
    }
	
  //---------------------------------------------------------------------------
  // Plot residuals
  //---------------------------------------------------------------------------
  double OM = par[0];
  double eOM= err[0];

  double OL = par[1];
  double eOL= err[1];
  double H  = par[2]; // This is really a nuisance parameter!

  double chi2 = 0;

  for(int i=0; i < NSN; i++)
    {
      MU_RES[i] = MU[i] - distanceModulus(Z[i], OM, OL, H);
      double chi = MU_RES[i] / MU_ERR[i]; 
      chi2 += chi*chi;
    }

  int ND = NSN-N; // number of degrees of freedom

  // // superimpose best fit
  // int NFIT = 100;
  // double zstep = 1.5/NFIT;
  // for(int i=0; i < NFIT; i++)
  //   {
  //     Z[i] = i*zstep;
  //     MU_FIT[i] = distanceModulus(Z[i], OM, OL, H);
  //   }
  // TGraphErrors* gfit = mkgraphErrors(NFIT, Z, MU_FIT, 0,
  // 				     "redshift z", "distance modulus #mu", 
  // 				     0, 1.5, 32.0, 48.0, kRed+1);
  // canvas->cd();
  // gfit->Draw("c same");


 

  TCanvas* cr = new TCanvas(string("fig_"+nom+"_residuals").c_str(), 
			    "SN1a", 
			    520, 10, 500, 500);

  TGraphErrors* gr = mkgraphErrors(NSN, Z, MU_RES, MU_ERR,
				   "redshift  z", "#Delta#mu = #mu_{i} - #mu", 
				   0, 1.5, -4.0, 4.0, kBlue+2);
  gr->SetLineWidth(1);
  gr->SetMarkerSize(0.2);

  cr->cd();
  gPad->SetGridy();
  gr->Draw("AP");

  Scribe* wrtext2 = new Scribe(0.25, 0.40);
  wrtext2->write("The Union2.1 Compilation");
  wrtext2->write("The Supernova Cosmology Project");
  wrtext2->write("http://supernova.lbl.gov/");
  wrtext2->write("Union/figures/SCPUnion2.1_mu_vs_z.txt");
  cr->SaveAs(".gif");
  cr->SaveAs(".pdf");

  cout << endl << endl;
  char record[2000];
  sprintf(record, 
	  "chi2/ND = %6.2f/%d\n"
	  "OM      = %6.2f +/- %-6.2f\n"
	  "OL      = %6.2f +/- %-6.2f\n"
	  "lifetime= %6.2f Gyr (assuming H0 = 70 km/s/Mpc)",
	  chi2, ND, OM, eOM, OL, eOL, lifetime(OM, OL, 70));
  cout << record << endl;
}


